/*
 * ToRegExpAlgebraic.cpp
 *
 *  Created on: 11. 2. 2014
 *	  Author: Tomas Pecka
 */

#include "ToRegExpAlgebraic.h"

#include <alphabet/Symbol.h>
#include <automaton/FSM/DFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <exception/AlibException.h>

#include "../../equations/RightRegularEquationSolver.h"
#include <regexp/unbounded/UnboundedRegExpElements.h>

namespace automaton {

namespace convert {

regexp::RegExp ToRegExpAlgebraic::convert(const automaton::Automaton& automaton) {
	return regexp::RegExp(getInstance().dispatch(automaton.getData()));
}

regexp::UnboundedRegExp ToRegExpAlgebraic::convert( const automaton::EpsilonNFA & automaton ) {
	equations::RightRegularEquationSolver solver;

	// initialize equations
	for( const auto & q : automaton.getStates( ) )
		solver.addSymbol( alphabet::Symbol( alphabet::LabeledSymbol( q.getName( ) ) ) );

	for( const auto & q : automaton.getStates( ) ) {
		if( automaton.getFinalStates( ).count( q ) > 0 )
			solver.addEquation( alphabet::Symbol( alphabet::LabeledSymbol ( q.getName( ) ) ), regexp::UnboundedRegExpEpsilon { } );
	}

	for( const auto & p : automaton.getSymbolTransitions() ) {
		for( const auto & q : p.second ) {
			solver.addEquation( alphabet::Symbol( alphabet::LabeledSymbol ( p.first.first.getName() ) ), alphabet::Symbol( alphabet::LabeledSymbol( q.getName() ) ), regexp::UnboundedRegExpSymbol { p.first.second } );
		}
	}

	for( const auto & p : automaton.getEpsilonTransitions() ) {
		for( const auto & q : p.second ) {
			solver.addEquation( alphabet::Symbol( alphabet::LabeledSymbol ( p.first.getName() ) ), alphabet::Symbol( alphabet::LabeledSymbol( q.getName() ) ), regexp::UnboundedRegExpEpsilon { } );
		}
	}

	return solver.solve( alphabet::Symbol( alphabet::LabeledSymbol (automaton.getInitialState().getName() ) ) );
}

auto ToRegExpAlgebraicEpsilonNFA = ToRegExpAlgebraic::RegistratorWrapper<regexp::UnboundedRegExp, automaton::EpsilonNFA>(ToRegExpAlgebraic::getInstance(), ToRegExpAlgebraic::convert);

regexp::UnboundedRegExp ToRegExpAlgebraic::convert( const automaton::MultiInitialStateNFA & automaton ) {
	equations::RightRegularEquationSolver solver;

	// initialize equations
	for( const auto & q : automaton.getStates( ) )
		solver.addSymbol( alphabet::Symbol( alphabet::LabeledSymbol( q.getName( ) ) ) );

	for( const auto & q : automaton.getStates( ) ) {
		if( automaton.getFinalStates( ).count( q ) > 0 )
			solver.addEquation( alphabet::Symbol( alphabet::LabeledSymbol ( q.getName( ) ) ), regexp::UnboundedRegExpEpsilon { } );
	}

	for( const auto & p : automaton.getTransitions() ) {
		for( const auto & q : p.second ) {
			solver.addEquation( alphabet::Symbol( alphabet::LabeledSymbol ( p.first.first.getName() ) ), alphabet::Symbol( alphabet::LabeledSymbol( q.getName() ) ), regexp::UnboundedRegExpSymbol { p.first.second } );
		}
	}

	regexp::UnboundedRegExpAlternation alternation;
	for(const auto& initialSymbol : automaton.getInitialStates() ) {
		// set symbol for which the solver will solve equation system
		alphabet::Symbol tmp( alphabet::LabeledSymbol ( initialSymbol.getName( ) ) );

		alternation.appendElement( solver.solve( tmp ).getRegExp() );
	}

	return regexp::UnboundedRegExp { alternation };
}

auto ToRegExpAlgebraicMultiInitialStateNFA = ToRegExpAlgebraic::RegistratorWrapper<regexp::UnboundedRegExp, automaton::MultiInitialStateNFA>(ToRegExpAlgebraic::getInstance(), ToRegExpAlgebraic::convert);

regexp::UnboundedRegExp ToRegExpAlgebraic::convert( const automaton::NFA & automaton ) {
	equations::RightRegularEquationSolver solver;

	// initialize equations
	for( const auto & q : automaton.getStates( ) )
		solver.addSymbol( alphabet::Symbol( alphabet::LabeledSymbol( q.getName( ) ) ) );

	for( const auto & q : automaton.getStates( ) ) {
		if( automaton.getFinalStates( ).count( q ) > 0 )
			solver.addEquation( alphabet::Symbol( alphabet::LabeledSymbol ( q.getName( ) ) ), regexp::UnboundedRegExpEpsilon { } );
	}

	for( const auto & p : automaton.getTransitions() ) {
		for( const auto & q : p.second ) {
			solver.addEquation( alphabet::Symbol( alphabet::LabeledSymbol ( p.first.first.getName() ) ), alphabet::Symbol( alphabet::LabeledSymbol( q.getName() ) ), regexp::UnboundedRegExpSymbol { p.first.second } );
		}
	}

	return solver.solve( alphabet::Symbol( alphabet::LabeledSymbol (automaton.getInitialState().getName() ) ) );
}

auto ToRegExpAlgebraicNFA = ToRegExpAlgebraic::RegistratorWrapper<regexp::UnboundedRegExp, automaton::NFA>(ToRegExpAlgebraic::getInstance(), ToRegExpAlgebraic::convert);

regexp::UnboundedRegExp ToRegExpAlgebraic::convert( const automaton::DFA & automaton ) {
	equations::RightRegularEquationSolver solver;

	// initialize equations
	for( const auto & q : automaton.getStates( ) )
		solver.addSymbol( alphabet::Symbol( alphabet::LabeledSymbol( q.getName( ) ) ) );

	for( const auto & q : automaton.getStates( ) ) {
		if( automaton.getFinalStates( ).count( q ) > 0 )
			solver.addEquation( alphabet::Symbol( alphabet::LabeledSymbol ( q.getName( ) ) ), regexp::UnboundedRegExpEpsilon { } );
	}

	for( const auto & p : automaton.getTransitions() ) {
		solver.addEquation( alphabet::Symbol( alphabet::LabeledSymbol ( p.first.first.getName() ) ), alphabet::Symbol( alphabet::LabeledSymbol( p.second.getName() ) ), regexp::UnboundedRegExpSymbol { p.first.second } );
	}

	return solver.solve( alphabet::Symbol( alphabet::LabeledSymbol (automaton.getInitialState().getName() ) ) );
}

auto ToRegExpAlgebraicDFA = ToRegExpAlgebraic::RegistratorWrapper<regexp::UnboundedRegExp, automaton::DFA>(ToRegExpAlgebraic::getInstance(), ToRegExpAlgebraic::convert);

} /* namespace convert */

} /* namespace automaton */
