/*
 * AutomataUnionEpsilonTransition.cpp
 *
 *  Created on: 20. 11. 2014
 *	  Author: Tomas Pecka
 */

#include "AutomataUnionEpsilonTransition.h"
#include <exception/AlibException.h>
#include "common/PairLabel.h"

#define AUTOMATON_FIRST  1
#define AUTOMATON_SECOND 2

namespace automaton {

namespace transform {

automaton::Automaton AutomataUnionEpsilonTransition::unification(const automaton::Automaton& first, const automaton::Automaton& second) {
	return getInstance().dispatch(first.getData(), second.getData());
}

automaton::EpsilonNFA AutomataUnionEpsilonTransition::unification(const automaton::EpsilonNFA& first, const automaton::EpsilonNFA& second) {
	std::set<automaton::State> states;
	for(const auto& q : first.getStates())
		states.insert(pairLabel(label::labelFrom(AUTOMATON_FIRST), q));
	for(const auto& q : second.getStates())
		states.insert(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	automaton::State q0 = automaton::createUniqueState(automaton::State("q0"), states);
	automaton::EpsilonNFA res(q0);

	for(const auto& a : first.getInputAlphabet())
		res.addInputSymbol(a);
	for(const auto& a : second.getInputAlphabet())
		res.addInputSymbol(a);

	for(const auto& q : states)
		res.addState(q);

	for(const auto& q : first.getFinalStates())
		res.addFinalState(pairLabel(label::labelFrom(AUTOMATON_FIRST), q));
	for(const auto& q : second.getFinalStates())
		res.addFinalState(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	res.addTransition(q0, pairLabel(label::labelFrom(AUTOMATON_FIRST), first.getInitialState()));
	res.addTransition(q0, pairLabel(label::labelFrom(AUTOMATON_SECOND), second.getInitialState()));

	for(const auto& t : first.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(pairLabel(label::labelFrom(AUTOMATON_FIRST), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_FIRST), q));

	for(const auto& t : second.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(pairLabel(label::labelFrom(AUTOMATON_SECOND), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	return res;
}

auto AutomataUnionEpsilonTransitionEpsilonNFA = AutomataUnionEpsilonTransition::RegistratorWrapper<automaton::EpsilonNFA, automaton::EpsilonNFA>(AutomataUnionEpsilonTransition::getInstance(), AutomataUnionEpsilonTransition::unification);

automaton::EpsilonNFA AutomataUnionEpsilonTransition::unification(const automaton::NFA& first, const automaton::NFA& second) {
	std::set<automaton::State> states;
	for(const auto& q : first.getStates())
		states.insert(pairLabel(label::labelFrom(AUTOMATON_FIRST), q));
	for(const auto& q : second.getStates())
		states.insert(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	automaton::State q0 = automaton::createUniqueState(automaton::State("q0"), states);
	automaton::EpsilonNFA res(q0);

	for(const auto& a : first.getInputAlphabet())
		res.addInputSymbol(a);
	for(const auto& a : second.getInputAlphabet())
		res.addInputSymbol(a);

	for(const auto& q : states)
		res.addState(q);

	for(const auto& q : first.getFinalStates())
		res.addFinalState(pairLabel(label::labelFrom(AUTOMATON_FIRST), q));
	for(const auto& q : second.getFinalStates())
		res.addFinalState(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	res.addTransition(q0, pairLabel(label::labelFrom(AUTOMATON_FIRST), first.getInitialState()));
	res.addTransition(q0, pairLabel(label::labelFrom(AUTOMATON_SECOND), second.getInitialState()));

	for(const auto& t : first.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(pairLabel(label::labelFrom(AUTOMATON_FIRST), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_FIRST), q));

	for(const auto& t : second.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(pairLabel(label::labelFrom(AUTOMATON_SECOND), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	return res;
}

auto AutomataUnionEpsilonTransitionNFA = AutomataUnionEpsilonTransition::RegistratorWrapper<automaton::EpsilonNFA, automaton::NFA>(AutomataUnionEpsilonTransition::getInstance(), AutomataUnionEpsilonTransition::unification);

automaton::EpsilonNFA AutomataUnionEpsilonTransition::unification(const automaton::DFA& first, const automaton::DFA& second) {
	std::set<automaton::State> states;
	for(const auto& q : first.getStates())
		states.insert(pairLabel(label::labelFrom(AUTOMATON_FIRST), q));
	for(const auto& q : second.getStates())
		states.insert(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	automaton::State q0 = automaton::createUniqueState(automaton::State("q0"), states);
	automaton::EpsilonNFA res(q0);

	for(const auto& a : first.getInputAlphabet())
		res.addInputSymbol(a);
	for(const auto& a : second.getInputAlphabet())
		res.addInputSymbol(a);

	for(const auto& q : states)
		res.addState(q);

	for(const auto& q : first.getFinalStates())
		res.addFinalState(pairLabel(label::labelFrom(AUTOMATON_FIRST), q));
	for(const auto& q : second.getFinalStates())
		res.addFinalState(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	res.addTransition(q0, pairLabel(label::labelFrom(AUTOMATON_FIRST), first.getInitialState()));
	res.addTransition(q0, pairLabel(label::labelFrom(AUTOMATON_SECOND), second.getInitialState()));

	for(const auto& t : first.getTransitions())
		res.addTransition(pairLabel(label::labelFrom(AUTOMATON_FIRST), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_FIRST), t.second));

	for(const auto& t : second.getTransitions())
		res.addTransition(pairLabel(label::labelFrom(AUTOMATON_SECOND), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_SECOND), t.second));

	return res;
}

auto AutomataUnionEpsilonTransitionDFA = AutomataUnionEpsilonTransition::RegistratorWrapper<automaton::EpsilonNFA, automaton::DFA>(AutomataUnionEpsilonTransition::getInstance(), AutomataUnionEpsilonTransition::unification);

} /* namespace transform */

} /* namespace automaton */
